# Test file for PSet #4
import operator
import math,numpy,random

rel_error = 1e-50

# erf,erfc from http://www.digitalmars.com/archives/cplusplus/3634.html (bottom of page)

def erf(x):
    if abs(x) > 2.2:
        # use continued fraction for large arguments
        return 1.0 - erfc(x)
    sum = x
    term = x
    xsqr = x*x
    j = 1
    while True:
        term *= xsqr/j
        sum -= term/(2*j+1)
        j += 1
        term *= xsqr/j
        sum += term/(2*j+1)
        j += 1
        if sum == 0 or abs(term/sum) <= rel_error:
            break
    return (2.0/math.sqrt(math.pi)) * sum

def erfc(x):
    if abs(x) < 2.2:
        # use series when small arguments
        return 1.0 - erf(x)
    if x < 0:
        # continued fraction only valid for x > 0
        return 2.0 - erfc(-x)
    a = 1; b = x            # last two convergent numerators
    c = x; d = x*x + 0.5    # last two convergent denominators
    q1 = a/c; q2 = b/d      # last two convergents (a/c and b/d)
    n = 1.0
    while True:
        a,b = (b,a*n + b*x)
        c,d = (d,c*n + d*x)
        n += 0.5
        q1,q2 = (q2,b/d)
        if abs(q1-q2)/q2 <= rel_error:
            break
    return (1.0/math.sqrt(math.pi))*math.exp(-x*x)*q2

def unit_normal_cdf(x):
    if x < 0:
        return 1 - unit_normal_cdf(-x)
    else:
        return 0.5 + 0.5*erf(x/math.sqrt(2))

# convert digital sample sequence into voltages, upsample, add noise
# result is numpy array
def transmit(message,samples_per_bit=4,nsigma=0,v0=0.0,v1=1.0,h=None):
    s = [v0]*(samples_per_bit*len(message))
    slen = len(s)
    for i in xrange(len(message)):
        if message[i] == '1' or message[i] == 1:
            j = i * samples_per_bit
            s[j:j + samples_per_bit] = [v1]*samples_per_bit

    if h is None:
        # construct an h with a wide-open eye, eg, [0.5, 0.25, 0.125, 0.125]
        h = [1.0/(2<<i) for i in xrange(samples_per_bit)]
        h[-1] += 1 - sum(h)  # adjust so that step reaches 1

    s = numpy.convolve(numpy.array(s),numpy.array(h))[:slen]

    if nsigma > 0:
        s = s + numpy.random.normal(0,nsigma,slen)

    return s

random.seed(2536038)
message = [random.randint(0,1) for i in xrange(100000)]
samples_per_bit = 4
channel_data = transmit(message,samples_per_bit=samples_per_bit,h=[.275,.25,.25,.225])
noisy_channel_data = transmit(message,samples_per_bit=samples_per_bit,nsigma=0.18,h=[.275,.25,.25,.225])

def test_sample_stats(sample_stats):
    stats = sample_stats(channel_data)
    expect = ((0.225,0.363,0.388),(0.025,0.263,0.355),(0.275,0.388,0.404),(0.500,0.500,0.500))
    error = False
    for i in xrange(samples_per_bit):
        smin,savg,sstd = stats[i]
        emin,eavg,estd = expect[i]
        if abs(smin - emin) > .001:
            print 'For sample time',i,'expected a min of',emin,'but got',smin
            error = True
        if abs(savg - eavg) > .001:
            print 'For sample time',i,'expected an average of',eavg,'but got',savg
            error = True
        if abs(sstd - estd) > .001:
            print 'For sample time',i,'expected a standard deviation of',estd,'but got',sstd
            error = True
    return None if error else 1

def test_bit_error_rate(bit_error_rate):
    message = [random.randint(0,1) for i in xrange(30)]
    xmessage = message[:]
    indices = (0,2,3,7,11,23,24,25,29)
    for index in indices:
        xmessage[index] = 1-message[index]
    ber = bit_error_rate(message,xmessage)
    eber = float(len(indices))/len(message)
    if ber != eber:
        print 'Called bit_error_rate, expected',eber,'got',ber
        print 'First arg: ',message
        print 'Second arg:',xmessage
        return None
    return 1

def test_compute_ber(compute_ber):
    for expect,sigma,vth,v0,v1,p0,p1 in ((.0227501,0.25,0.5,0,1,.5,.5),
                                         (.0237315,0.25,0.4,0,1,1./3,2./3)):
        got = compute_ber(sigma,vth=vth,v0=v0,v1=v1,p0=p0,p1=p1)
        if abs(got - expect) > 1e-6:
            print 'For compute_ber(%g,vth=%g,v0=%g,v1=%g,p0=%g,p1=%g):' % \
                  (sigma,vth,v0,v1,p0,p1)
            print '  expected = %g, got %g' % (expect,got)

# return True if the binary sequence contains an even number of 1's
def even_parity(seq):
    return reduce(operator.xor,seq,0) == 0

# construct a (nrows*ncols+nrows+ncols,nrows*ncols,3) codeword
def codeword(data,nrows,ncols):
    result = data[:]
    # compute row parity bits
    for r in xrange(nrows):
        result.append(int(not even_parity(data[r*ncols:(r+1)*ncols])))
    # compute column parity bits
    for c in xrange(ncols):
        result.append(int(not even_parity(data[c:len(data):ncols])))
    return result

# convert integer to binary list of specified size (lsb first)
def int2bin(c,size=8):
    return [(c >> i) % 2 for i in xrange(size)]

def test_correct_errors(correct_errors):
    # test the decoder on a particular value, returns True if failed
    def run_test(test,nrows,ncols,expected,msg):
        result = correct_errors(test[:],nrows,ncols)
        if result != expected:
            print "Error detected while testing",msg
            print "Test code word:",test
            print "Expected:",expected
            print "Received:",result
            return True
        return False

    # test all 256 data values for 8 data bits
    #for i in xrange(256):
        #data = int2bin(i)
        #if run_test(codeword(data,2,4),2,4,data,"valid codewords"): return
    data = int2bin(75)
    if run_test(codeword(data,2,4),2,4,data,"valid codewords"): return

    # test all single-bit errors (should be corrected)
    for count in range(10):
        good = codeword(int2bin((count*37 + 15) % 256),4,2)
        for i in xrange(14):
            bad = good[:]
            bad[i] ^= 1
            if run_test(bad,4,2,good[0:8],"correctable single-bit errors (bit %d)" % i): return

    print 'Tests completed successfully'

