import numpy, os

# convert digital sample sequence into voltages; result is numpy array.
# samples_per_bit can be a float (eg, 8.1) to model clock
# drift in the transmitter.
# NOTE JKW; NO DEFAULT VALUES FOR SAMPLES_PER_BIT, causes confusion
def bits_to_digitized_samples(bits,samples_per_bit,ntraining,repeat = 1):
    """ generate sequence of voltage samples:
          bits: binary message sequence
          samples_per_bit: number of samples for each message bit
          ntraining: number of bits of training seq (0,1,...)
          repeat: how many times to repeat whole shebang
    """
    # Preallocate samples list and default to all v0's
    samples = [0] * int((ntraining + len(bits))*samples_per_bit)
    index = 0
    for i in xrange(ntraining):
        if i % 2 == 1:
            start = int(index)
            end = int(index + samples_per_bit)
            samples[start:end] = [1]*(end-start)
        index = index + samples_per_bit
    for i in xrange(len(bits)):
        if bits[i] != 0:
            start = int(index)
            end = int(index + samples_per_bit)
            samples[start:end] = [1]*(end-start)
        index = index + samples_per_bit
    
    return(samples * repeat)

# convert nbits of int into bit sequences, lsb first
def int_to_bits(n,nbits=8):
    return [(n >> i) & 1 for i in xrange(nbits)]

def print_bits(bits):
    return " ".join(["".join(['%d'%bits[i] for i in xrange(j,min(j+4,len(bits)))])
                     for j in xrange(0,len(bits)+3,4)])

# tester for task 1.  Type is 'same', 'fast', 'slow'
# set debug to True to see what actually happened
def task1_test(f,type,debug=False,nbits=256):
    # construct a test message: a sequence of n-bit integers
    n = 4
    bits = [0]*((1<<n) * n)
    for i in xrange(1 << n):
        start = n*i
        bits[start:start+n] = int_to_bits(i,nbits=n)
    if nbits/16 > 1:
        bits = bits * (nbits/16)
    if debug:
        print "message bits:",print_bits(bits)

    # convert it into digitized samples
    if type == 'same':
        nominal_samples_per_bit = 11
        actual_samples_per_bit = 11
    elif type == 'fast':
        nominal_samples_per_bit = 8
        actual_samples_per_bit = 8.1
    elif type == 'slow':
        nominal_samples_per_bit = 5
        actual_samples_per_bit = 4.8

    # see if user's function can handle it
    ntraining = nominal_samples_per_bit * 2
    if debug:
        print "number of training bits:",ntraining

    print "Testing with %d samples per bit, type is %s." % (nominal_samples_per_bit,type)

    digitized_samples = bits_to_digitized_samples(bits,actual_samples_per_bit,ntraining)
    if debug:
        print "digitized samples:",digitized_samples

    result = f(digitized_samples,nominal_samples_per_bit)
    if debug:
        print "training bits returned by function:",print_bits(result[:ntraining])
        print "message bits returned by function:",print_bits(result[ntraining:])

    # check each message bit
    # skip training bits since they may be wrong until receiver gets in sync
    for i in xrange(len(bits)):
        if bits[i] != result[i+ntraining]:
            print "returned bit sequence first differs from expected at message bit",i
            break
    else:
        print 'test passed'
    print ''

# encode list of integers into sequence of bits
#  prefix = sequence of bits that appear before data bits
#  postfix = sequence of bits that appear after data bits
#  nbits = number of data bits to encode (LSB first)
def encode_data(data,prefix=[1],postfix=[0],nbits=8):
    # start with prefix
    result = prefix[:]
    # encode data LSB first
    result.extend(int_to_bits(data,nbits=8))
    # end with postfix bits
    result.extend(postfix)
    return result

# encode numpy array of bits in to an array of encoded bits of
# nbits-length words with prefix and postfix
# prefix = sequence of bits that appear before data bits
# postfix = sequence of bits that appear after data bits
# nbits = number of data bits to encode
def encode_bits(data,prefix=[1],postfix=[0],nbits=8):
    assert len(data) % nbits == 0, "Data length incorrect size!"
    num_words = len(data)/8
    pad = 5
    word_len = nbits + len(prefix)+len(postfix)
    enc_bits = numpy.zeros(num_words*word_len  + 2 * pad, int)
    offset = pad
    for i in xrange(num_words):
        word = prefix[:] + list(data[nbits*i:nbits*(i+1)]) + postfix[:]
        enc_bits[offset:offset+word_len] = word
        offset = offset + word_len
    return enc_bits

# encode string.  See encode_data for details...
def encode_string(s,prefix=[1],postfix=[0],nbits=8):
    result = [0]*5   # start with some zeros
    for ch in s:
        result.extend(encode_data(ord(ch),prefix=prefix,postfix=postfix,nbits=nbits))
    result += [0]*5   # end with some zeros
    return result

# convert bit sequence (lsb first) to an int
def bits_to_int(bits):
    result = 0
    for i in xrange(len(bits)):
        result += bits[i] * (1 << i)
    return int(result)

# 8b/10 encoder

# two encoder tables: 5b/6b and 3b/4b
# each entry is ((int,int),(int,int))
# RD is -1: use tuple[0]; if RD is +1: use tuple[1]
# code is tuple[x][0]
# new RD is tuple[x][1]

table_5b_6b = [
    ((0x39,1), (0x06,-1)),  # D.00
    ((0x2E,1), (0x11,-1)),  # D.01
    ((0x2D,1), (0x12,-1)),  # D.02
    ((0x23,-1), (0x23,1)),  # D.03
    ((0x2B,1), (0x14,-1)),  # D.04
    ((0x25,-1), (0x25,1)),  # D.05
    ((0x26,-1), (0x26,1)),  # D.06
    ((0x07,-1), (0x38,1)),  # D.07
    ((0x27,1), (0x18,-1)),  # D.08
    ((0x29,-1), (0x29,1)),  # D.09
    ((0x2A,-1), (0x2A,1)),  # D.10
    ((0x0B,-1), (0x0B,1)),  # D.11
    ((0x2C,-1), (0x2C,1)),  # D.12
    ((0x0D,-1), (0x0D,1)),  # D.13
    ((0x0E,-1), (0x0E,1)),  # D.14
    ((0x3A,1), (0x05,-1)),  # D.15
    ((0x36,1), (0x09,-1)),  # D.16
    ((0x31,-1), (0x31,1)),  # D.17
    ((0x32,-1), (0x32,1)),  # D.18
    ((0x13,-1), (0x13,1)),  # D.19
    ((0x34,-1), (0x34,1)),  # D.20
    ((0x15,-1), (0x15,1)),  # D.21
    ((0x16,-1), (0x16,1)),  # D.22
    ((0x17,1), (0x28,-1)),  # D.23
    ((0x33,1), (0x0C,-1)),  # D.24
    ((0x19,-1), (0x19,1)),  # D.25
    ((0x1A,-1), (0x1A,1)),  # D.26
    ((0x1B,1), (0x24,-1)),  # D.27
    ((0x1C,-1), (0x1C,1)),  # D.28
    ((0x1D,1), (0x22,-1)),  # D.29
    ((0x1E,1), (0x21,-1)),  # D.30
    ((0x35,1), (0x0A,-1)),  # D.31
    ]

# index by (byte >> 5) & 0x7
# use D.x.A7 when
#   if RD is -1, and x is 17, 18, 20
#   if RD is +1, and x 11, 13, 14
table_3b_4b = [
    ((0xD,1), (0x2,-1)),  # D.x.0
    ((0x9,-1), (0x9,1)),  # D.x.1
    ((0xA,-1), (0xA,1)),  # D.x.2
    ((0x3,-1), (0xC,1)),  # D.x.3
    ((0xB,1), (0x4,-1)),  # D.x.4
    ((0x5,-1), (0x5,1)),  # D.x.5
    ((0x6,-1), (0x6,1)),  # D.x.6
    ((0x7,1), (0x8,-1)),  # D.x.P7
    ((0xE,1), (0x1,-1)),  # D.x.A7
    ]

# 7-bit sync to use for RD==-1, RD==+1:
sync = (0x7C,0x03)

# (sync,updated rd) = encode_sync_8b10b(rd)
def encode_sync_8b10b(rd):
    if rd == -1:
        return (sync[0],1)
    else:
        return (sync[1],-1)

# (10b,updated rd) = encode_8b_10b(8b,rd)
def encode_data_8b10b(ch,rd):
    # encode low-order 6 bits
    x = ch & 0x1F
    x_6b = table_5b_6b[x]
    xcode,rd = x_6b[0] if rd == -1 else x_6b[1]
    #print 'x_6b=(%x,%x), rd=%d' % (x_6b[0],x_6b[1],rd)

    # encode high-order 3 bits
    y = (ch >> 5) & 0x7
    y_4b = table_3b_4b[y]
    if (rd == -1 and y == 7 and x in (17,18,20)) or \
       (rd ==  1 and y == 7 and x in (11,13,14)):
        y_4b = table_3b_4b[8]
    ycode,rd = y_4b[0] if rd == -1 else y_4b[1]
    #print 'y_4b=(%x,%x), rd=%d' % (y_4b[0],y_4b[1],rd)

    return ((ycode << 6) + xcode,rd)

# padding at end of packets
padding = (7,21,2,14,11)

# encode string with 8b/10b inserting syncs every so often
def encode_string_8b10b(s,sync_interval = 16):
    result = [0,1]*5   # start with training sequence
    packet = 0
    count = 0
    rd = -1   # running disparity starts at -1
    for ch in s:
        if (count % sync_interval) == 0:
            sync,rd = encode_sync_8b10b(rd)
            result.extend(int_to_bits(sync,7))
            #print 'end of sync at',len(result)
        code,rd = encode_data_8b10b(ord(ch),rd)
        #print '%d: code(%s = D.%02x.%x) = %s' % (len(result),ch,ord(ch) & 0x1F, ord(ch) >> 5,int_to_bits(code,10))
        result.extend(int_to_bits(code,10))
        count += 1
        if (count % sync_interval) == 0:
            # add some "random" padding at end of packet
            result += [0,1]*padding[packet % len(padding)]
            packet += 1
            #print '==========='

    return result

# build the reverse lookup table by encoding each character
# with both possible running disparities and then filling
# in the appropriate entries
table_10b_8b = [None] * 1024
for ch in xrange(256):
    for rd in (-1,1):
        code = encode_data_8b10b(ch,rd)[0]
        #print "0x%x (D.%02x.%x) encodes as 0x%03x when rd=%d" % (ch,ch & 0x1F, ch >> 5,code,rd)
        if not(table_10b_8b[code] is None or table_10b_8b[code]==ch):
            print "Oops, duplicate entry in table_10b_8b: ch=0x%x, code=0x%x, prev entry=0x%x" % (ch,code,table_10b_8b[code])
        table_10b_8b[code] = ch

short_message = """
test
"""
long_message = """
Fourscore and seven years ago our fathers brought forth on this
continent a new nation, conceived in liberty, and dedicated to the
proposition that all men are created equal.

Now we are engaged in a great civil war, testing whether that nation,
or any nation, so conceived and so dedicated, can long endure. We are
met on a great battle-field of that war. We have come to dedicate a
portion of that field as a final resting place for those who here
gave their lives that that nation might live. It is altogether fitting
and proper that we should do this.

But, in a larger sense, we cannot dedicate... we cannot
consecrate... we cannot hallow... this ground. The brave men, living
and dead, who struggled here, have consecrated it far above our poor
power to add or detract. The world will little note nor long remember
what we say here, but it can never forget what they did here. It is
for us, the living, rather, to be dedicated here to the unfinished work
which they who fought here have thus far so nobly advanced. It is
rather for us to be here dedicated to the great task remaining before
us... that from these honored dead we take increased devotion to that
cause for which they gave the last full measure of devotion; that we
here highly resolve that these dead shall not have died in vain; that
this nation, under God, shall have a new birth of freedom; and that
government of the people, by the people, for the people, shall not
perish from the earth.

November 19, 1863
"""

# tester for task 3
def task2_test(f,message):
    "testjig for PSet #2, Task #2 -- expects your receive function and a string as arguments"
    data = encode_string_8b10b(message)
    result = f(data)
    if result != message:
        print 'expected "%s", got "%s"' % (message,result)
    else:
        print 'received message is "%s"' % result
